@@ -5,6 +5,7 @@ import logging
 import re
 
 from subprocess import TimeoutExpired
+from typing import List
 from retry import retry
 from .download_bitwarden import DownloadBitwarden
 
@@ -44,10 +45,7 @@ class Bitwarden(object):
             **os.environ,
         }
 
-        if (
-            self.bitwarden_credentials["client_id"]
-            and self.bitwarden_credentials["client_secret"]
-        ):
+        if self.bitwarden_credentials["client_id"] and self.bitwarden_credentials["client_secret"]:
             env["BW_CLIENTID"] = str(self.bitwarden_credentials["client_id"])
             env["BW_CLIENTSECRET"] = str(self.bitwarden_credentials["client_secret"])
             logging.info("Using provided client_id and client_secret")
@@ -56,9 +54,9 @@ class Bitwarden(object):
 
         try:
             if not env["BW_CLIENTID"] or not env["BW_CLIENTSECRET"]:
-                raise Exception("Empty client_id or client_secret!")
-        except KeyError:
-            raise Exception("Empty client_id or client_secret!")
+                raise ValueError("Empty client_id or client_secret!")
+        except KeyError as e:
+            raise ValueError("Empty client_id or client_secret!") from e
 
         try:
             return subprocess.run(
@@ -84,7 +82,7 @@ class Bitwarden(object):
                 env=env,
             )
 
-    @retry(BitwardenServerException, delay=5, tries=3)
+    @retry((TimeoutExpired, BitwardenServerException), delay=5, tries=3)
     def bitwarden_login(self):
         """
         Performs login opeartion via BitWarden CLI
@@ -105,26 +103,24 @@ class Bitwarden(object):
             )
 
             if "Invalid master password" in bitwarden_app.stderr:
-                logging.error("STDOUT: " + bitwarden_app.stdout)
-                logging.error("STDERR: " + bitwarden_app.stderr)
-                raise Exception("Invalid master password!")
+                logging.error(f"STDOUT: {bitwarden_app.stdout}")
+                logging.error(f"STDERR: {bitwarden_app.stderr}")
+                raise ValueError("Invalid master password!")
 
             if not re.search(r".{80,}", bitwarden_app.stdout):
-                logging.error("STDOUT: " + bitwarden_app.stdout)
-                logging.error("STDERR: " + bitwarden_app.stderr)
-                raise Exception("No session key returned after login!")
+                logging.error(f"STDOUT: {bitwarden_app.stdout}")
+                logging.error(f"STDERR: {bitwarden_app.stderr}")
+                raise RuntimeError("No session key returned after login!")
 
             self.session_key = bitwarden_app.stdout
         else:
-            logging.error("STDOUT: " + bitwarden_app.stdout)
-            logging.error("STDERR: " + bitwarden_app.stderr)
+            logging.error(f"STDOUT: {bitwarden_app.stdout}")
+            logging.error(f"STDERR: {bitwarden_app.stderr}")
 
             if "invalid_client" in bitwarden_app.stderr:
-                raise Exception("Invalid bitwarden client_id or client_secret!")
+                raise ValueError("Invalid bitwarden client_id or client_secret!")
 
-            raise BitwardenServerException(
-                "Bitwarden Server error during login! Retrying..."
-            )
+            raise BitwardenServerException("Bitwarden Server error during login! Retrying...")
 
     def get_credentials(self, user_credentials_name):
         """
@@ -134,6 +130,7 @@ class Bitwarden(object):
         self.get_data(user_credentials_name)
         return self.data
 
+    @retry((TimeoutExpired, BitwardenServerException), delay=5, tries=3)
     def get_data(self, data):
         """
         Core method
@@ -168,23 +165,20 @@ class Bitwarden(object):
         )
 
         if not bitwarden_app.stdout:
-            raise Exception("Empty items list! Probably some issue with login!")
+            logging.error(f"STDERR: {bitwarden_app.stderr}")
+            raise RuntimeError("Empty result for 'bw list items' command! This user has no items associated")
 
         bitwarden_items = json.loads(bitwarden_app.stdout)
+        items_not_found: List[str] = []
         for credentials_key, credentials_name in data.items():
             for bw_item in bitwarden_items:
                 if credentials_name == bw_item["name"]:
-                    self.data[credentials_key] = {}
-                    self.data[credentials_key]["login"] = bw_item["login"]["username"]
-                    self.data[credentials_key]["password"] = bw_item["login"][
-                        "password"
-                    ]
-                    if "uris" in bw_item["login"]:
-                        self.data[credentials_key]["url"] = bw_item["login"]["uris"][0][
-                            "uri"
-                        ]
-                    else:
-                        self.data[credentials_key]["url"] = ""
+                    self.data[credentials_key] = {
+                        "login": bw_item["login"]["username"],
+                        "password": bw_item["login"]["password"],
+                        "url": bw_item["login"]["uris"][0]["uri"] if "uris" in bw_item["login"] else "",
+                    }
+
                     if bw_item["login"]["totp"] is None:
                         self.data[credentials_key]["otp"] = ""
                     else:
@@ -199,12 +193,20 @@ class Bitwarden(object):
                     if "fields" in bw_item:
                         for field in bw_item["fields"]:
                             self.data[credentials_key][field["name"]] = field["value"]
-        if len(self.data) != len(data):
-            raise Exception(
-                "Invalid bitwarden collection or key name or no access to collection for this user!"
+
+                    break
+            else:
+                items_not_found.append(credentials_name)
+
+        if items_not_found:
+            logging.error(f"STDOUT: {bitwarden_app.stdout}")
+            logging.error(f"STDERR: {bitwarden_app.stderr}")
+            raise BitwardenServerException(
+                "Invalid bitwarden collection or key name or no access to collection for this user! Items not found "
+                f"include {', '.join(items_not_found)}"
             )
 
-    @retry(TimeoutExpired, delay=5, tries=3)
+    @retry((TimeoutExpired, BitwardenServerException), delay=5, tries=3)
     def get_attachment(self, item_name, file_name, output_folder=os.getcwd()):
         """
         Downloads attachment file from particular item to current working directory
@@ -222,12 +224,8 @@ class Bitwarden(object):
             f = open("att.txt", "r")
             assert f.read() == "secret text\n"
         """
-        if (
-            not isinstance(item_name, str)
-            or not isinstance(file_name, str)
-            or not isinstance(output_folder, str)
-        ):
-            raise Exception("item_name / file_name / output_folder should be strings!")
+        if not isinstance(item_name, str) or not isinstance(file_name, str) or not isinstance(output_folder, str):
+            raise TypeError("item_name / file_name / output_folder should be strings!")
 
         if not os.path.exists(output_folder):
             os.mkdir(output_folder)
@@ -249,19 +247,15 @@ class Bitwarden(object):
         )
 
         if bitwarden_app.stderr:
-            logging.error("STDOUT: " + bitwarden_app.stdout)
-            logging.error("STDERR: " + bitwarden_app.stderr)
+            logging.error(f"STDOUT: {bitwarden_app.stdout}")
+            logging.error(f"STDERR: {bitwarden_app.stderr}")
+            if "A 404 error occurred while downloading the attachment" in bitwarden_app.stderr:
+                raise BitwardenServerException("404 HTTP from bitwarden server occurred!")
             if "More than one result was found" in bitwarden_app.stderr:
-                raise Exception(
-                    "More than one result for item name was found! Name should be unique!"
-                )
+                raise ValueError("More than one result for item name was found! Name should be unique!")
             if "Not found" in bitwarden_app.stderr:
-                raise Exception(
-                    "Cannot find bitwarden item '{}'! Check the name".format(
-                        str(item_name)
-                    )
-                )
-            raise Exception("Unknown error during items obtaining!")
+                raise ValueError(f"Cannot find bitwarden item '{str(item_name)}'! Check the name")
+            raise RuntimeError("Unknown error during items obtaining!")
 
         bitwarden_items = json.loads(bitwarden_app.stdout)
         item_id = str(bitwarden_items["id"])
@@ -279,12 +273,12 @@ class Bitwarden(object):
         )
 
         if bitwarden_app.stderr:
-            logging.error("STDOUT: " + bitwarden_app.stdout)
-            logging.error("STDERR: " + bitwarden_app.stderr)
+            logging.error(f"STDOUT: {bitwarden_app.stdout}")
+            logging.error(f"STDERR: {bitwarden_app.stderr}")
 
             if "was not found" in bitwarden_app.stderr:
-                raise Exception("Attachment was not found!")
+                raise ValueError("Attachment was not found!")
 
-            raise Exception("Unknown error during attachment downloading!")
+            raise RuntimeError("Unknown error during attachment downloading!")
 
-        print("Attachment '{}' downloaded sucessfully!".format(file_name))
+        print(f"Attachment '{file_name}' downloaded sucessfully!")
